
/*************************************************************************
	
	file created on:	2002/08/30   19:33
	filename: 			Bmp.cpp
	author:				Andreas Hartl

		visit http://www.runicsoft.com for updates and more information

	purpose:	functions to load raw bmp data,
	                      to save raw bmp data,
						  to convert RGB data to raw bmp data,
						  to convert raw bmp data to RGB data
						  and to use the WinAPI to select
							a bitmap into a device context

	file updated on 2010/09/13
		in the 8 years since i first wrote this the windows file functions
		have changed their input from char* to LPCTSTR. 
		Updated this in the code here

**************************************************************************/

//note: this file is a simplified version of the original

#include <stdio.h>       // for fopen, etc.
#include <string.h>       // for memset

#include <iostream> //for cout
#include "bmp.h"

#define BI_RGB 0

using namespace std;

/*
typedef struct tagBITMAPFILEHEADER 
{
   WORD    bfType;        // must be 'BM' 
   DWORD   bfSize;        // size of the whole .bmp file
   WORD    bfReserved1;   // must be 0
   WORD    bfReserved2;   // must be 0
   DWORD   bfOffBits;     
} BITMAPFILEHEADER;
//total size is 14B, 7 WORDS


typedef struct tagBITMAPINFOHEADER
{
   DWORD  biSize;            // size of the structure
   LONG   biWidth;           // image width
   LONG   biHeight;          // image height
   WORD   biPlanes;          // bitplanes
   WORD   biBitCount;        // resolution 
   DWORD  biCompression;     // compression
   DWORD  biSizeImage;       // size of the image
   LONG   biXPelsPerMeter;   // pixels per meter X
   LONG   biYPelsPerMeter;   // pixels per meter Y
   DWORD  biClrUsed;         // colors used
   DWORD  biClrImportant;    // important colors
} BITMAPINFOHEADER;
//total size is 40B, 20 WORDS
*/




/*******************************************************************
BYTE* ConvertRGBToBMPBuffer ( BYTE* Buffer, int width, 
		int height, long* newsize )


This function takes as input an array of RGB values, it's width
and height.
The buffer gets then transformed to an array that can be used
to write to a windows bitmap file. The size of the array
is returned in newsize, the array itself is the
return value of the function.
Both input and output buffers must be deleted by the
calling function.

The input buffer is expected to consist of width * height
RGB triplets. Thus the total size of the buffer is taken as
width * height * 3.

The function then transforms this buffer so that it can be written 
to a windows bitmap file:
First the RGB triplets are converted to BGR.
Then the buffer is swapped around since .bmps store
images uside-down.
Finally the buffer gets DWORD ( 32bit ) aligned, 
meaning that each scanline ( 3 * width bytes ) gets
padded with 0x00 bytes up to the next DWORD boundary


*******************************************************************/

BYTE* ConvertRGBToBMPBuffer ( BYTE* Buffer, int width, int height, long* newsize )
{

	// first make sure the parameters are valid
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;

	// now we have to find with how many bytes
	// we have to pad for the next DWORD boundary	

	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )     // DWORD = 4 bytes
		padding++;
	// get the padded scanline width
	int psw = scanlinebytes + padding;
	
	// we can already store the size of the new padded buffer
	*newsize = height * psw;

	// and create new buffer
	BYTE* newbuf = new BYTE[*newsize];
	
	// fill the buffer with zero bytes then we dont have to add
	// extra padding zero bytes later on
	memset ( newbuf, 0, *newsize );

	// now we loop trough all bytes of the original buffer, 
	// swap the R and B bytes and the scanlines
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < 3 * width; x+=3 )
		{
			bufpos = y * 3 * width + x;     // position in original buffer
			newpos = ( height - y - 1 ) * psw + x;           // position in padded buffer

			newbuf[newpos] = Buffer[bufpos+2];       // swap r and b
			newbuf[newpos + 1] = Buffer[bufpos + 1]; // g stays
			newbuf[newpos + 2] = Buffer[bufpos];     // swap b and r
		}

	return newbuf;
}

BYTE* ConvertGRAY8ToBMPBuffer ( BYTE* Buffer, int width, int height, long* newsize )
{

	// first make sure the parameters are valid
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;

	// now we have to find with how many bytes
	// we have to pad for the next DWORD boundary	

	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )     // DWORD = 4 bytes
		padding++;
	// get the padded scanline width
	int psw = scanlinebytes + padding;
	
	// we can already store the size of the new padded buffer
	*newsize = height * psw;

	// and create new buffer
	BYTE* newbuf = new BYTE[*newsize];
	
	// fill the buffer with zero bytes then we dont have to add
	// extra padding zero bytes later on
	memset ( newbuf, 0, *newsize );

	// now we loop trough all bytes of the original buffer, 
	// swap the R and B bytes and the scanlines
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < width; x+=1 )
		{
			bufpos = y * width + x;     // position in original buffer
			newpos = ( height - y - 1 ) * psw + x *3;           // position in padded buffer

			newbuf[newpos] = Buffer[bufpos]; 
			newbuf[newpos + 1] = Buffer[bufpos];
			newbuf[newpos + 2] = Buffer[bufpos];
		}

	return newbuf;
}


/***************************************************************
BYTE* ConvertBMPToRGBBuffer ( BYTE* Buffer, 
		int width, int height )

This function takes as input the data array
from a bitmap and its width and height.
It then converts the bmp data into an RGB array.
The calling function must delete both the input
and output arrays.
The size of the returned array will be 
width * height * 3
On error the returb value is NULL, else the
RGB array.


The Buffer is expected to be the exact data read out
from a .bmp file.  
The function will swap it around, since images
are stored upside-down in bmps.
The BGR triplets from the image data will
be converted to RGB.
And finally the function removes padding bytes.
The returned arraay consits then of
width * height RGB triplets.

*****************************************************************/

BYTE* ConvertBMPToRGBBuffer ( BYTE* Buffer, int width, int height, BYTE* newbuf )
{
	// first make sure the parameters are valid
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;

	// find the number of padding bytes
		
	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )     // DWORD = 4 bytes
		padding++;
	// get the padded scanline width
	int psw = scanlinebytes + padding;

	// create new buffer
	//BYTE* newbuf = new BYTE[width*height*3];
	
	// now we loop trough all bytes of the original buffer, 
	// swap the R and B bytes and the scanlines
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < 3 * width; x+=3 )
		{
			newpos = y * 3 * width + x;     
			bufpos = ( height - y - 1 ) * psw + x;

			newbuf[newpos] = Buffer[bufpos + 2];       
			newbuf[newpos + 1] = Buffer[bufpos+1]; 
			newbuf[newpos + 2] = Buffer[bufpos];     
		}

	return newbuf;
}

BYTE* ConvertBMPToGRAY8Buffer ( BYTE* Buffer, int width, int height, BYTE* newbuf )
{
	// first make sure the parameters are valid
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;

	// find the number of padding bytes
		
	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )     // DWORD = 4 bytes
		padding++;
	// get the padded scanline width
	int psw = scanlinebytes + padding;

	// create new buffer
	//BYTE* newbuf = new BYTE[width*height];
	
	// now we loop trough all bytes of the original buffer, 
	// swap the R and B bytes and the scanlines
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < width; x+=1 )
		{
			newpos = y * width + x;     
			bufpos = ( height - y - 1 ) * psw + 3*x;

			//0.2126 R + 0.7152 G + 0.0722 B
			//poradi je BGR u BMP
			float tmp=0.0722f*(float)Buffer[bufpos  ]
					  + 0.7152f*(float)Buffer[bufpos+1]
					  + 0.2126f*(float)Buffer[bufpos+2];     
			newbuf[newpos]  = (BYTE)tmp;
		}

	return newbuf;
}




/***************************************************************
bool SaveBMP ( BYTE* Buffer, int width, int height, 
		long paddedsize, LPCTSTR bmpfile )

Function takes a buffer of size <paddedsize> 
and saves it as a <width> * <height> sized bitmap 
under the supplied filename.
On error the return value is false.

***************************************************************/

bool SaveBMP ( BYTE* Buffer, int width, int height, long paddedsize, const LPCTSTR bmpfile )
{
	unsigned short tmp[20];

	
	// fill the fileheader with data
	tmp[0]=0x4d42;
	tmp[1]=(54+paddedsize)& 0x0000ffff;
	tmp[2]=(54+paddedsize)>>16;
	tmp[3]=tmp[4]=0; //reserved
	tmp[5]=54;tmp[6]=0; //OffBits
	/*
	bmfh.bfType = 0x4d42;       // 0x4d42 = 'BM'
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + paddedsize;
	bmfh.bfOffBits = 0x36;		// number of bytes to start of bitmap bits
	*/

	// now we open the file to write to
	FILE *file=fopen(bmpfile, "wb");
	if ( file == NULL ) return false;
	
	// write file header
	fwrite(&tmp,1,14,file);

	tmp[0]=40; tmp[1]=0;
	tmp[2]=width& 0x0000ffff;
	tmp[3]=width >>16;
	tmp[4]=height& 0x0000ffff;
	tmp[5]=height >>16;
	tmp[6]=1;
	tmp[7]=24;
	tmp[8]=BI_RGB; tmp[9]=0;
	tmp[10]=tmp[11]=0;
	tmp[12]=0x0ec4; tmp[13]=0;
	tmp[14]=0x0ec4; tmp[15]=0;
	tmp[16]=tmp[17]=tmp[18]=tmp[19]=0;
	/*
	info.biSize = sizeof(BITMAPINFOHEADER);
	info.biWidth = width;
	info.biHeight = height;
	info.biPlanes = 1;			// we only have one bitplane
	info.biBitCount = 24;		// RGB mode is 24 bits
	info.biCompression = BI_RGB;	
	info.biSizeImage = 0;		// can be 0 for 24 bit images
	info.biXPelsPerMeter = 0x0ec4;     // paint and PSP use this values
	info.biYPelsPerMeter = 0x0ec4;     
	info.biClrUsed = 0;			// we are in RGB mode and have no palette
	info.biClrImportant = 0;    // all colors are important
	*/

	// write infoheader
	fwrite(&tmp,1,40,file);

	// write image data
	fwrite(Buffer,1,paddedsize, file);
	
	// and close file
	fclose(file);
	
	// fill the infoheader

	return true;
}

/*******************************************************************
BYTE* LoadBMP ( int* width, int* height, long* size 
		LPCTSTR bmpfile )

The function loads a 24 bit bitmap from bmpfile, 
stores it's width and height in the supplied variables
and the whole size of the data (padded) in <size>
and returns a buffer of the image data 

On error the return value is NULL. 

  NOTE: make sure you [] delete the returned array at end of 
		program!!!
*******************************************************************/

BYTE* LoadBMP ( int* width, int* height, long* size, const LPCTSTR bmpfile )
{
	unsigned short tmp[20];

	// open file to read from
	FILE *file=fopen(bmpfile, "rb");
	if ( file == NULL ) return NULL;
	
	
	// read file header
	fread(&tmp,1,14,file);
	
	// check if file is actually a bmp
	//if ( bmpheader.bfType != 0x4d42 ) // 0x4d42 = 'BM'
	if (tmp[0] != 0x4d42) 
	{
		fclose(file);
		return NULL;
	}

	long skip=(long)tmp[5] + ((long)tmp[6]<<16);
	*size = (long)tmp[1] + ((long)tmp[2]<<16) - skip;

	cout << "data size je " << *size << "B" << endl;
	cout << "data zacinaji na offsetu " << skip << "B" << endl;

	//read bitmap info
	fread(&tmp,1,40,file);

	// get image measurements
	//*width   = bmpinfo.biWidth;
	//*height  = abs ( bmpinfo.biHeight );
	*width   = (int)tmp[2] + ((int)tmp[3]<<16);
	*height  = (int)tmp[4] + ((int)tmp[5]<<16);

	cout << "rozmery jsou " << *width << " x " << *height << endl;
	cout << "komprese je " << tmp[8] << " (ma byt 0)" << endl;

	// check if bmp is uncompressed
	if ( tmp[8] != BI_RGB )
	{
		fclose(file);
		return NULL;
	}

	// check if we have 24 bit bmp
	//if ( bmpinfo.biBitCount != 24 )
	if ( tmp[7] != 24 )
	{
		fclose(file);
		return NULL;
	}
	

	// create buffer to hold the data
	//*size = bmpheader.bfSize - bmpheader.bfOffBits;

	BYTE* Buffer = new BYTE[ *size ];
	if (Buffer == NULL) { cout << "nedostal jsem pamet!\n"; return NULL; }

	// move file pointer to start of bitmap data
	fseek(file,skip,SEEK_SET);
	// read bmp data
	fread(Buffer,1,*size,file);

	// everything successful here: close file and return buffer
	
	fclose(file);


	return Buffer;
}

/* priklad
void TestBMPCopy2(LPCTSTR input, LPCTSTR output)
{
	int x, y;
	long s, s2;
	BYTE* a = LoadBMP ( &x, &y, &s, input );
	//BYTE* b = ConvertBMPToRGBBuffer ( a, x, y );
	//BYTE* c = ConvertRGBToBMPBuffer ( b, x, y, &s2 );	
	BYTE* b = ConvertBMPToGRAY8Buffer ( a, x, y );
	BYTE* c = ConvertGRAY8ToBMPBuffer ( b, x, y, &s2 );	
	SaveBMP ( c, x, y, s2, output );
	delete [] a;
	delete [] b;
	delete [] c;
}
*/
