/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * FILE: filters.cc
 *
 * Image filtering
 *
 * Petr Mejzlk (mejzlik@fi.muni.cz) 2000
 * Honza Huben (xhubeny@fi.muni.cz) 2004
 */

#ifdef __GNUG__
#pragma implementation
#endif

#include "filters.h"

namespace i3d {
  using namespace std;
  using std::abs;


// -----------------------------
// Vlado's FIR Gauss starts here
// -----------------------------

template <class VOXEL> void EstimateGaussKernel(struct Separable3dFilter <VOXEL > &filter,
						   const VOXEL sigmaX,
						   const VOXEL sigmaY,
						   const VOXEL sigmaZ,
						   const float widthX,
						   const float widthY,
						   const float widthZ)
{
    filter.DisposeData();
    filter.xData = filter.yData = filter.zData = NULL;

    filter.xLength = 0;
    filter.yLength = 0;
    filter.zLength = 0;

    filter.xDiv = 1;
    filter.yDiv = 1;
    filter.zDiv = 1;

    if ((sigmaX <= 0.0) && (sigmaY <= 0.0) && (sigmaZ <= 0.0))
    {
    	std::cerr << "EstimateGaussKernel(): Warning! Requested creation of an empty filter!\n";

    	return;
    }

    int i;
    VOXEL value;

    if (sigmaX > 0) {
    	const int halfsize = static_cast <int>(ceil(widthX * sigmaX));
    	const int length = halfsize * 2 + 1;

	filter.xData = new VOXEL[length];
    	filter.xLength = length;

    	for (i = 0; i <= halfsize; ++i)
    	{
        	value = exp(static_cast <VOXEL> ((-(i * i)) / (2.0 * sigmaX * sigmaX)));
        	filter.xData[halfsize + i] = value;
        	filter.xData[halfsize - i] = value;
    	}

    	filter.xDiv = filter.xData[halfsize];
    	for (i = 1; i <= halfsize; ++i)
    	{
        	filter.xDiv += filter.xData[i + halfsize] * 2;
    	}
    }

    if (sigmaY > 0) {
    	const int halfsize = static_cast <int>(ceil(widthY * sigmaY));
    	const int length = halfsize * 2 + 1;

	filter.yData = new VOXEL[length];
    	filter.yLength = length;

    	for (i = 0; i <= halfsize; ++i)
    	{
        	value = exp(static_cast <VOXEL> ((-(i * i)) / (2.0 * sigmaY * sigmaY)));
        	filter.yData[halfsize + i] = value;
        	filter.yData[halfsize - i] = value;
    	}

    	filter.yDiv = filter.yData[halfsize];
    	for (i = 1; i <= halfsize; ++i)
    	{
        	filter.yDiv += filter.yData[i + halfsize] * 2;
    	}
    }

    if (sigmaZ > 0) {
    	const int halfsize = static_cast <int>(ceil(widthZ * sigmaZ));
    	const int length = halfsize * 2 + 1;

	filter.zData = new VOXEL[length];
    	filter.zLength = length;

    	for (i = 0; i <= halfsize; ++i)
    	{
        	value = exp(static_cast <VOXEL> ((-(i * i)) / (2.0 * sigmaZ * sigmaZ)));
        	filter.zData[halfsize + i] = value;
        	filter.zData[halfsize - i] = value;
    	}

    	filter.zDiv = filter.zData[halfsize];
    	for (i = 1; i <= halfsize; ++i)
    	{
        	filter.zDiv += filter.zData[i + halfsize] * 2;
    	}
    }
}

template <class VOXEL> void PrepareZeroBoundaryForFilter(struct Separable3dFilter <VOXEL > const &filter, BorderPadding<VOXEL> &b)
{
    b.DisposeData();
    b.xLength = 0;
    b.yLength = 0;
    b.zLength = 0;

    if (filter.xLength > 0) {
        b.xLength = filter.xLength / 2;
	b.xDataPre=new VOXEL[b.xLength];
	b.xDataPost=b.xDataPre;
	for (long i=0; i < b.xLength; ++i) b.xDataPre[i]=0;
    }
    if (filter.yLength > 0) {
        b.yLength = filter.yLength / 2;
	b.yDataPre=new VOXEL[b.yLength];
	b.yDataPost=b.yDataPre;
	for (long i=0; i < b.yLength; ++i) b.yDataPre[i]=0;
    }
    if (filter.zLength > 0) {
        b.zLength = filter.zLength / 2;
	b.zDataPre=new VOXEL[b.zLength];
	b.zDataPost=b.zDataPre;
	for (long i=0; i < b.zLength; ++i) b.zDataPre[i]=0;
    }
}

template <class VOXEL> void GaussFIR(i3d::Image3d <VOXEL> const &In, i3d::Image3d <VOXEL> &Out,
				const VOXEL Sigma, const float Width)
{
	Separable3dFilter<VOXEL> f;
	BorderPadding<VOXEL> b;

	const VOXEL SigmaX=(In.GetSizeX() > 1)? Sigma : 0;
	const VOXEL SigmaY=(In.GetSizeY() > 1)? Sigma : 0;
	const VOXEL SigmaZ=(In.GetSizeZ() > 1)? Sigma : 0;

	EstimateGaussKernel<VOXEL>(f,SigmaX,SigmaY,SigmaZ,Width,Width,Width);
	PrepareZeroBoundaryForFilter<VOXEL>(f,b);

	SeparableConvolution<VOXEL>(In,Out,f,b);
}

template <class VOXEL> void GaussFIR(i3d::Image3d <VOXEL> &img,
				const VOXEL Sigma, const float Width)
{
	Image3d<VOXEL> in(img);
	GaussFIR<VOXEL>(in,img,Sigma,Width);
}

template <class VOXEL> void GaussFIR(i3d::Image3d <VOXEL> const &In, i3d::Image3d <VOXEL> &Out,
				const VOXEL SigmaX, const VOXEL SigmaY, const VOXEL SigmaZ,
				const float WidthX, const float WidthY, const float WidthZ)
{
	Separable3dFilter<VOXEL> f;
	BorderPadding<VOXEL> b;

	EstimateGaussKernel<VOXEL>(f,SigmaX,SigmaY,SigmaZ,WidthX,WidthY,WidthZ);
	PrepareZeroBoundaryForFilter<VOXEL>(f,b);

	SeparableConvolution<VOXEL>(In,Out,f,b);
}

template <class VOXEL> void GaussFIR(i3d::Image3d <VOXEL> &img,
				const VOXEL SigmaX, const VOXEL SigmaY, const VOXEL SigmaZ,
				const float WidthX, const float WidthY, const float WidthZ)
{
	Image3d<VOXEL> in(img);
	GaussFIR<VOXEL>(in,img,SigmaX,SigmaY,SigmaZ,WidthX,WidthY,WidthZ);
}
// ---------------------------
// Vlado's FIR Gauss ends here
// ---------------------------


// -----------------------------
// Vlado's IIR Gauss starts here
// -----------------------------

namespace { //internal functions, start of the unnamed namespace (containing a local function)
/**
 \ingroup gf_helper
 Compute the IIR filter coefficients for Gauss filter with zero mean and given \e Sigma.
 This is a part of IIR version of Gauss filter.

 \note This is related to [Vliet, Young and Ginkel, 2002].

 \param[in] sigma	the desired sigma of the filter
 \param[out] B		the FIR part coefficient
 \param[out] b1,b2,b3	the IIR part coefficients
*/
void ComputeBsFromSigma(const double sigma, double &B, double &b1, double &b2, double &b3) {
	//the code was borrowed from Geusebrock 2002 ( //http://staff.science.uva.nl/~mark/anigauss.c ) in great part
	double q, qsq;
	double scale;

	/* initial values */
	double m0 = 1.16680, m1 = 1.10783, m2 = 1.40586;
	double m1sq = m1*m1, m2sq = m2*m2;

	/* calculate q */
	if(sigma < 3.556)
		q = -0.2568 + 0.5784 * sigma + 0.0561 * sigma * sigma;
	else
		q = 2.5091 + 0.9804 * (sigma - 3.556);

	qsq = q*q;

	/* calculate scale, and b[0,1,2,3] */
	scale = (m0 + q) * (m1sq + m2sq + 2*m1*q + qsq);
	b1 = -q * (2*m0*m1 + m1sq + m2sq + (2*m0 + 4*m1)*q + 3*qsq) / scale;
	b2 = qsq * (m0 + 2*m1 + 3*q) / scale;
	b3 = - qsq * q / scale;
    
	/* calculate B */
	B = (m0 * (m1sq + m2sq))/scale; B*=B;
}

/**
 \ingroup gf_helper
 Estimate the boundary settings for Young and van Vliet's shaped IIR filters according
 to the [Triggs and Sdika, 2006]. It is applicable only to ApplyIIRGauss2() and ApplyIIRGauss3() functions.

 \b Literature: B. Triggs and M. Sdika. Boundary conditions for Young-van Vliet recursive filtering.
 In \b Signal \b processing 54, 5. 2006.

 \param[in] b1,b2,b3	IIR filter parameters, should be the same for both forward and backward directions
 \param[out] M		pointer to boundary conditions adjusting matrix

 \note \e M should be initialized beforehand and should be able to hold at least 9 \c FLOAT values.
*/
template <typename FLOAT>
void SuggestIIRBoundaries(const FLOAT b1,const FLOAT b2,const FLOAT b3, FLOAT *M) {
	FLOAT scale = 1/((1+b1-b2+b3)*(1-b1-b2-b3)*(1+b2+(b1-b3)*b3));

	M[0] = scale*(-b3*b1+1-b3*b3-b2);
	M[1] = scale*(b3+b1)*(b2+b3*b1);
	M[2] = scale*b3*(b1+b3*b2);
	M[3] = scale*(b1+b3*b2);
	M[4] = -scale*(b2-1)*(b2+b3*b1);
	M[5] = -scale*b3*(b3*b1+b3*b3+b2-1);
	M[6] = scale*(b3*b1+b2+b1*b1-b2*b2);
	M[7] = scale*(b1*b2+b3*b2*b2-b1*b3*b3-b3*b3*b3-b3*b2+b3);
	M[8] = scale*b3*(b1+b3*b2);
}

} // end of the unnamed namespace (containing a local function)


//the following three macros enable debuging outputs (very very detailed) of
//the core of the GaussIIR() and the SeparableConvolution() functions

//#define DEBUG_CORE_X
//#define DEBUG_CORE_Y
//#define DEBUG_CORE_Z

//#define REMOVE_REMAINDER_FORWARD(val)	      val=(val < 0.0005)? 0 : val;	//removes
#define REMOVE_REMAINDER_FORWARD(val)   					//leaves
//#define REMOVE_REMAINDER_BACKWARD(val)      val=(val < 0.0005)? 0 : val;	//removes
#define REMOVE_REMAINDER_BACKWARD(val)  					//leaves

// see header file for docs; 2002 version
template <class VOXEL>
  void GaussIIR(Image3d <VOXEL> &img, const VOXEL Sigma) {

    //input/output image dimensions
    const signed long x_size = (signed long) img.GetWidth();
    const signed long y_size = (signed long) img.GetHeight();
    const signed long z_size = (signed long) img.GetNumSlices();
    const signed long s_size = x_size * y_size;     //just a helper variables (timesavers)
    const signed long x_2size = 2*x_size;
    const signed long x_3size = 3*x_size;
    const signed long s_2size = 2*s_size;
    const signed long s_3size = 3*s_size;

    if ((x_size > 1) && (x_size < 4)) {
    	std::cerr << "GaussIIR(): Input image too small along x-axis, min is 4px.\n";
	return;
    }
    if ((y_size > 1) && (y_size < 4)) {
    	std::cerr << "GaussIIR(): Input image too small along y-axis, min is 4px.\n";
	return;
    }
    if ((z_size > 1) && (z_size < 4)) {
    	std::cerr << "GaussIIR(): Input image too small along z-axis, min is 4px.\n";
	return;
    }

    //setting the filter shape
    double b1D,b2D,b3D,BD;
    double S=static_cast<double>(Sigma);
    ComputeBsFromSigma(S,BD,b1D,b2D,b3D);

    VOXEL b1=static_cast<VOXEL>(b1D),
    	  b2=static_cast<VOXEL>(b2D),
  	  b3=static_cast<VOXEL>(b3D),
	  B=static_cast<VOXEL>(BD);

    VOXEL *M=new VOXEL[9];		//boundary handling matrix
    if (M == NULL) {
    	std::cerr << "GaussIIR(): No memory left for boundary matrix.\n";
	return;
    }
    SuggestIIRBoundaries(-b1,-b2,-b3,M);
    const VOXEL denom=1+b1+b2+b3;	//boundary handling...

#ifdef I3D_DEBUG
    std::cout << "GaussIIR: input image of size " << x_size << "x" << y_size << "x" << z_size << " voxels\n";
    std::cout << "GaussIIR: B=" << B << ", b1=" << b1 << ", b2=" << b2 << ", b3=" << b3 << std::endl;
#endif
    //temporary variables
    signed long i = 0, c = 0;

    VOXEL *wrk = img.GetFirstVoxelAddr();

    // ---- X AXIS ----
    if (img.GetSizeX() > 1)
    {
        //x-axis filter is present...
#ifdef I3D_DEBUG
        std::cout << "GaussIIR: convolving in x-axis...\n";
#endif
        for (signed long rows = 0; rows < y_size * z_size; rows++)
        {
            //no need to shift wrk pointer because x-lines are sequentially in memory

	    //boundary handling...
	    const VOXEL bnd=*wrk/denom;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_X
	    std::cout << "GaussIIR:   forward: prefix starts, ";
#endif
	    *wrk=*wrk - bnd*(b1+b2+b3); //i=0
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    ++wrk;

	    *wrk=*wrk  - *(wrk-1)*b1 - bnd*(b2+b3); //i=1
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    ++wrk;

	    *wrk=*wrk  - *(wrk-1)*b1 - *(wrk-2)*b2 - bnd*b3; //i=2
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    ++wrk;

#ifdef DEBUG_CORE_X
	    std::cout << "convolutions of i=0,1,2 (row=" << rows << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < (x_size-1); i++)
            {
#ifdef DEBUG_CORE_X
                std::cout << "GaussIIR:   forward: infix starts, ";
#endif
		*wrk=*wrk  - *(wrk-1)*b1 - *(wrk-2)*b2 - *(wrk-3)*b3;
		REMOVE_REMAINDER_FORWARD(*wrk)
#ifdef DEBUG_CORE_X
                std::cout << "convolution of " << i << " (row=" << rows << ") is finished\n";
#endif
                ++wrk;
            }

	    ++wrk;
        }

	wrk=img.GetVoxelAddr(img.GetImageSize()-1); //useless

        for (signed long rows = 0; rows < y_size * z_size; rows++)
        {
            //no need to shift wrk pointer because x-lines are sequentially in memory

	    VOXEL up=*wrk/denom; //a copy, btw: this equals to u_plus in (15) in [Triggs and Sdika, 2006]
	    
	    *wrk=*wrk  - *(wrk-1)*b1 - *(wrk-2)*b2 - *(wrk-3)*b3; //finishing the forward convolution
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    //the existence of (wrk-3) is guaranteed by the minimum image size test at the begining of this routine

	    //boundary handling...
	    const VOXEL bnd1_=*wrk-up, //new *wrk values after the entire forward convolution are used now
			bnd2_=*(wrk-1)-up,
			bnd3_=*(wrk-2)-up;
	    up/=denom; //this equals to v_plus in (15) in [Triggs and Sdika, 2006]

			*wrk=(M[0]*bnd1_ + M[1]*bnd2_ + M[2]*bnd3_ +up) *B;
	    const VOXEL bnd2=(M[3]*bnd1_ + M[4]*bnd2_ + M[5]*bnd3_ +up) *B;
	    const VOXEL bnd3=(M[6]*bnd1_ + M[7]*bnd2_ + M[8]*bnd3_ +up) *B;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_X
	    std::cout << "GaussIIR:   backward: prefix starts, ";
#endif
	    //*wrk*=B; //i=0
	    REMOVE_REMAINDER_BACKWARD(*wrk)
	    --wrk;

	    //*wrk=(*wrk)*B  - *(wrk+1)*b1 - bnd2*B*b2 - bnd3*B*b3; //i=1
	    *wrk=(*wrk)*B  - *(wrk+1)*b1 - bnd2*b2 - bnd3*b3; //i=1
	    REMOVE_REMAINDER_BACKWARD(*wrk)
	    --wrk;

	    //*wrk=(*wrk)*B  - *(wrk+1)*b1 - *(wrk+2)*b2 - bnd2*B*b3; //i=2
	    *wrk=(*wrk)*B  - *(wrk+1)*b1 - *(wrk+2)*b2 - bnd2*b3; //i=2
	    REMOVE_REMAINDER_BACKWARD(*wrk)
	    --wrk;

#ifdef DEBUG_CORE_X
	    std::cout << "convolutions of i=" << x_size-1 << "," << x_size-2 << "," << x_size-3 \
	    	      << " (row=" << rows << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < x_size; i++) //growing of i is unimportant in this case
            {
#ifdef DEBUG_CORE_X
                std::cout << "GaussIIR:   backward: infix starts, ";
#endif
		*wrk=(*wrk)*B  - *(wrk+1)*b1 - *(wrk+2)*b2 - *(wrk+3)*b3;
		REMOVE_REMAINDER_BACKWARD(*wrk)
#ifdef DEBUG_CORE_X
                std::cout << "convolution of " << x_size-1-i << " (row=" << rows << ") is finished\n";
#endif
                --wrk;
            }
        }

	wrk=img.GetFirstVoxelAddr(); //important
    }

    // ---- Y AXIS ----
    if (img.GetSizeY() > 1)
    {
        //y-axis filter is present...
        VOXEL *wrk2 = wrk;

#ifdef I3D_DEBUG
        std::cout << "GaussIIR: convolving in y-axis...\n";
#endif
	//storage of original input values before any y-axis related convolution
	VOXEL *bnd=new VOXEL[x_size];

        for (signed long slices = 0; slices < z_size; slices++)
        {
            wrk = wrk2 + (slices * s_size);

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Y
	    std::cout << "GaussIIR:   forward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //boundary handling...

	    	*(wrk+c)=*(wrk+c) - *(bnd+c)*(b1+b2+b3); //i=0
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += x_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-x_size)*b1 - *(bnd+c)*(b2+b3); //i=1
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += x_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-x_size)*b1 - *(wrk+c-x_2size)*b2 - *(bnd+c)*b3; //i=2
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += x_size;

#ifdef DEBUG_CORE_Y
	    std::cout << "convolutions of i=0,1,2 (slice=" << slices << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < (y_size-1); i++)
            {
#ifdef DEBUG_CORE_Y
                std::cout << "GaussIIR:   forward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c)  - *(wrk+c-x_size)*b1 - *(wrk+c-x_2size)*b2 - *(wrk+c-x_3size)*b3;
		    REMOVE_REMAINDER_FORWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Y
                std::cout << "convolution of " << i << " (slice=" << slices << ") is finished\n";
#endif
                wrk += x_size;
            }

            wrk += x_size;
        }

	wrk2=img.GetVoxelAddr(img.GetImageSize()) -x_size; //important

	VOXEL *bnd2=new VOXEL[x_size];
	VOXEL *bnd3=new VOXEL[x_size];
	VOXEL *vp=new VOXEL[x_size]; //to store v_plus in (15) in [Triggs and Sdika, 2006]

	VOXEL *Bnd2=new VOXEL[x_size];
	VOXEL *Bnd3=new VOXEL[x_size];

        for (signed long slices = 0; slices < z_size; slices++)
        {
            wrk = wrk2 - (slices * s_size);

	    //copy the voxel's value at the end of y-line and finish the forward convolution
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //will be used *bnd instead of *up, just for this occassion

		//finishes the forward convolution
		*(wrk+c)=*(wrk+c) - *(wrk+c-x_size)*b1 - *(wrk+c-x_2size)*b2 - *(wrk+c-x_3size)*b3;
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }

	    for (c = 0; c < x_size; c++) *(vp+c)=*(bnd+c)/denom;
	    for (c = 0; c < x_size; c++) *(bnd3+c)=*(wrk+c-x_2size)-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd2+c)=*(wrk+c-x_size )-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd +c)=*(wrk+c        )-*(bnd+c);

	    //for (c = 0; c < x_size; c++) *(wrk +c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd2+c)=( *(bnd+c)*M[3] + *(bnd2+c)*M[4] + *(bnd3+c)*M[5] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd3+c)=( *(bnd+c)*M[6] + *(bnd2+c)*M[7] + *(bnd3+c)*M[8] + *(vp+c) ) *B;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Y
	    std::cout << "GaussIIR:   backward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(wrk+c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= x_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+x_size)*b1 - *(Bnd2+c)*b2 - *(Bnd3+c)*b3; //i=1
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= x_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+x_size)*b1 - *(wrk+c+x_2size)*b2 - *(Bnd2+c)*b3; //i=2
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= x_size;

#ifdef DEBUG_CORE_Y
	    std::cout << "convolutions of i=" << y_size-1 << "," << y_size-2 << "," << y_size-3 \
	    	      << " (slice=" << slices << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < y_size; i++)
            {
#ifdef DEBUG_CORE_Y
                std::cout << "GaussIIR:   backward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c) * B  - *(wrk+c+x_size)*b1 - *(wrk+c+x_2size)*b2 - *(wrk+c+x_3size)*b3;
	            REMOVE_REMAINDER_BACKWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Y
                std::cout << "convolution of " << y_size-1-i << " (slice=" << slices << ") is finished\n";
#endif
                wrk -= x_size;
            }
        }

	//release allocated memory
	delete[] bnd;
	delete[] bnd2;
	delete[] bnd3;
	delete[] vp;
	delete[] Bnd2;
	delete[] Bnd3;

	wrk=img.GetFirstVoxelAddr(); //important
    }

    // ---- Z AXIS ----
    if (img.GetSizeZ() > 1)
    {
        //z-axis filter is present...
        VOXEL *wrk2 = wrk;

#ifdef I3D_DEBUG
        std::cout << "GaussIIR: convolving in z-axis...\n";
#endif
	//storage of original input values before any y-axis related convolution
	VOXEL *bnd=new VOXEL[x_size];

        for (signed long cols = 0; cols < y_size; cols++)
        {
            wrk = wrk2 + (cols * x_size);

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Z
	    std::cout << "GaussIIR:   forward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //boundary handling...

	    	*(wrk+c)=*(wrk+c) - *(bnd+c)*(b1+b2+b3); //i=0
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += s_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-s_size)*b1 - *(bnd+c)*(b2+b3); //i=1
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += s_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-s_size)*b1 - *(wrk+c-s_2size)*b2 - *(bnd+c)*b3; //i=2
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += s_size;

#ifdef DEBUG_CORE_Z
	    std::cout << "convolutions of i=0,1,2 (col=" << cols << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < (z_size-1); i++)
            {
#ifdef DEBUG_CORE_Z
                std::cout << "GaussIIR:   forward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c)  - *(wrk+c-s_size)*b1 - *(wrk+c-s_2size)*b2 - *(wrk+c-s_3size)*b3;
		    REMOVE_REMAINDER_FORWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Z
                std::cout << "convolution of " << i << " (col=" << cols << ") is finished\n";
#endif
                wrk += s_size;
            }

            wrk += s_size;
        }

	wrk2=img.GetVoxelAddr(img.GetImageSize()) -s_size;

	VOXEL *bnd2=new VOXEL[x_size];
	VOXEL *bnd3=new VOXEL[x_size];
	VOXEL *vp=new VOXEL[x_size]; //to store v_plus in (15) in [Triggs and Sdika, 2006]

	VOXEL *Bnd2=new VOXEL[x_size];
	VOXEL *Bnd3=new VOXEL[x_size];

        for (signed long cols = 0; cols < y_size; cols++)
        {
            wrk = wrk2 + (cols * x_size);

	    //copy the voxel's value at the end of z-line and finish the forward convolution
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //will be used *bnd instead of *up, just for this occassion

		//finishes the forward convolution
		*(wrk+c)=*(wrk+c) - *(wrk+c-s_size)*b1 - *(wrk+c-s_2size)*b2 - *(wrk+c-s_3size)*b3;
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }

	    for (c = 0; c < x_size; c++) *(vp+c)=*(bnd+c)/denom;
	    for (c = 0; c < x_size; c++) *(bnd3+c)=*(wrk+c-s_2size)-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd2+c)=*(wrk+c-s_size )-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd +c)=*(wrk+c        )-*(bnd+c);

	    //for (c = 0; c < x_size; c++) *(wrk +c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd2+c)=( *(bnd+c)*M[3] + *(bnd2+c)*M[4] + *(bnd3+c)*M[5] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd3+c)=( *(bnd+c)*M[6] + *(bnd2+c)*M[7] + *(bnd3+c)*M[8] + *(vp+c) ) *B;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Z
	    std::cout << "GaussIIR:   backward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(wrk+c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= s_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+s_size)*b1 - *(Bnd2+c)*b2 - *(Bnd3+c)*b3; //i=1
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= s_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+s_size)*b1 - *(wrk+c+s_2size)*b2 - *(Bnd2+c)*b3; //i=2
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= s_size;

#ifdef DEBUG_CORE_Z
	    std::cout << "convolutions of i=" << z_size-1 << "," << z_size-2 << "," << z_size-3 \
	    	      << " (col=" << cols << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < z_size; i++)
            {
#ifdef DEBUG_CORE_Z
                std::cout << "GaussIIR:   backward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c) * B  - *(wrk+c+s_size)*b1 - *(wrk+c+s_2size)*b2 - *(wrk+c+s_3size)*b3;
	            REMOVE_REMAINDER_BACKWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Z
                std::cout << "convolution of " << z_size-1-i << " (col=" << cols << ") is finished\n";
#endif
                wrk -= s_size;
            }
        }

	//release allocated memory
	delete[] bnd;
	delete[] bnd2;
	delete[] bnd3;
	delete[] vp;
	delete[] Bnd2;
	delete[] Bnd3;
    }

    delete[] M;				//release boundary handling matrix
}

template <class VOXEL, class PRECISION> void GrayGaussIIR(Image3d<VOXEL> &img, const PRECISION Sigma) 
{
	Image3d<PRECISION> tmp;
	GrayToFloat(img, tmp);
	GaussIIR(tmp, Sigma);
	FloatToGrayNoWeight(tmp, img);
}

// see header file for docs; 2002 version
template <class VOXEL>
  void GaussIIR(Image3d <VOXEL> &img, const VOXEL SigmaX, const VOXEL SigmaY, const VOXEL SigmaZ) {

    //input/output image dimensions
    const signed long x_size = (signed long) img.GetWidth();
    const signed long y_size = (signed long) img.GetHeight();
    const signed long z_size = (signed long) img.GetNumSlices();
    const signed long s_size = x_size * y_size;     //just a helper variables (timesavers)
    const signed long x_2size = 2*x_size;
    const signed long x_3size = 3*x_size;
    const signed long s_2size = 2*s_size;
    const signed long s_3size = 3*s_size;

    if ((x_size > 1) && (x_size < 4)) {
    	std::cerr << "GaussIIR(): Input image too small along x-axis, min is 4px.\n";
	return;
    }
    if ((y_size > 1) && (y_size < 4)) {
    	std::cerr << "GaussIIR(): Input image too small along y-axis, min is 4px.\n";
	return;
    }
    if ((z_size > 1) && (z_size < 4)) {
    	std::cerr << "GaussIIR(): Input image too small along z-axis, min is 4px.\n";
	return;
    }

    //setting the filter shape
    double b1D,b2D,b3D,BD;
    VOXEL B,b1,b2,b3;

    VOXEL *M=new VOXEL[9];		//boundary handling matrix
    if (M == NULL) {
    	std::cerr << "GaussIIR(): No memory left for boundary matrix.\n";
	return;
    }

#ifdef I3D_DEBUG
    std::cout << "GaussIIR: input image of size " << x_size << "x" << y_size << "x" << z_size << " voxels\n";
#endif
    //temporary variables
    signed long i = 0, c = 0;

    VOXEL *wrk = img.GetFirstVoxelAddr();

    // ---- X AXIS ----
    if ((img.GetSizeX() > 1) && (SigmaX >= 1.0))
    {
	double S=static_cast<double>(SigmaX);
	ComputeBsFromSigma(S,BD,b1D,b2D,b3D);

	b1=static_cast<VOXEL>(b1D);
	b2=static_cast<VOXEL>(b2D);
	b3=static_cast<VOXEL>(b3D);
	B=static_cast<VOXEL>(BD);

	SuggestIIRBoundaries(-b1,-b2,-b3,M);
	const VOXEL denom=1+b1+b2+b3;	//boundary handling...

        //x-axis filter is present...
#ifdef I3D_DEBUG
        std::cout << "GaussIIR: convolving in x-axis...\n";
#endif
#ifdef I3D_DEBUG
	std::cout << "GaussIIR: B=" << B << ", b1=" << b1 << ", b2=" << b2 << ", b3=" << b3 << std::endl;
#endif
        for (signed long rows = 0; rows < y_size * z_size; rows++)
        {
            //no need to shift wrk pointer because x-lines are sequentially in memory

	    //boundary handling...
	    const VOXEL bnd=*wrk/denom;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_X
	    std::cout << "GaussIIR:   forward: prefix starts, ";
#endif
	    *wrk=*wrk - bnd*(b1+b2+b3); //i=0
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    ++wrk;

	    *wrk=*wrk  - *(wrk-1)*b1 - bnd*(b2+b3); //i=1
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    ++wrk;

	    *wrk=*wrk  - *(wrk-1)*b1 - *(wrk-2)*b2 - bnd*b3; //i=2
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    ++wrk;

#ifdef DEBUG_CORE_X
	    std::cout << "convolutions of i=0,1,2 (row=" << rows << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < (x_size-1); i++)
            {
#ifdef DEBUG_CORE_X
                std::cout << "GaussIIR:   forward: infix starts, ";
#endif
		*wrk=*wrk  - *(wrk-1)*b1 - *(wrk-2)*b2 - *(wrk-3)*b3;
		REMOVE_REMAINDER_FORWARD(*wrk)
#ifdef DEBUG_CORE_X
                std::cout << "convolution of " << i << " (row=" << rows << ") is finished\n";
#endif
                ++wrk;
            }

	    ++wrk;
        }

	wrk=img.GetVoxelAddr(img.GetImageSize()-1); //useless

        for (signed long rows = 0; rows < y_size * z_size; rows++)
        {
            //no need to shift wrk pointer because x-lines are sequentially in memory

	    VOXEL up=*wrk/denom; //a copy, btw: this equals to u_plus in (15) in [Triggs and Sdika, 2006]
	    
	    *wrk=*wrk  - *(wrk-1)*b1 - *(wrk-2)*b2 - *(wrk-3)*b3; //finishing the forward convolution
	    REMOVE_REMAINDER_FORWARD(*wrk)
	    //the existence of (wrk-3) is guaranteed by the minimum image size test at the begining of this routine

	    //boundary handling...
	    const VOXEL bnd1_=*wrk-up, //new *wrk values after the entire forward convolution are used now
			bnd2_=*(wrk-1)-up,
			bnd3_=*(wrk-2)-up;
	    up/=denom; //this equals to v_plus in (15) in [Triggs and Sdika, 2006]

			*wrk=(M[0]*bnd1_ + M[1]*bnd2_ + M[2]*bnd3_ +up) *B;
	    const VOXEL bnd2=(M[3]*bnd1_ + M[4]*bnd2_ + M[5]*bnd3_ +up) *B;
	    const VOXEL bnd3=(M[6]*bnd1_ + M[7]*bnd2_ + M[8]*bnd3_ +up) *B;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_X
	    std::cout << "GaussIIR:   backward: prefix starts, ";
#endif
	    //done already few lines above, i=0
	    REMOVE_REMAINDER_BACKWARD(*wrk)
	    --wrk;

	    *wrk=(*wrk)*B  - *(wrk+1)*b1 - bnd2*b2 - bnd3*b3; //i=1
	    REMOVE_REMAINDER_BACKWARD(*wrk)
	    --wrk;

	    *wrk=(*wrk)*B  - *(wrk+1)*b1 - *(wrk+2)*b2 - bnd2*b3; //i=2
	    REMOVE_REMAINDER_BACKWARD(*wrk)
	    --wrk;

#ifdef DEBUG_CORE_X
	    std::cout << "convolutions of i=" << x_size-1 << "," << x_size-2 << "," << x_size-3 \
	    	      << " (row=" << rows << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < x_size; i++) //growing of i is unimportant in this case
            {
#ifdef DEBUG_CORE_X
                std::cout << "GaussIIR:   backward: infix starts, ";
#endif
		*wrk=(*wrk)*B  - *(wrk+1)*b1 - *(wrk+2)*b2 - *(wrk+3)*b3;
		REMOVE_REMAINDER_BACKWARD(*wrk)
#ifdef DEBUG_CORE_X
                std::cout << "convolution of " << x_size-1-i << " (row=" << rows << ") is finished\n";
#endif
                --wrk;
            }
        }

	wrk=img.GetFirstVoxelAddr(); //important
    }

    // ---- Y AXIS ----
    if ((img.GetSizeY() > 1) && (SigmaY >= 1.0))
    {
	double S=static_cast<double>(SigmaY);
	ComputeBsFromSigma(S,BD,b1D,b2D,b3D);

	b1=static_cast<VOXEL>(b1D);
	b2=static_cast<VOXEL>(b2D);
	b3=static_cast<VOXEL>(b3D);
	B=static_cast<VOXEL>(BD);

	SuggestIIRBoundaries(-b1,-b2,-b3,M);
	const VOXEL denom=1+b1+b2+b3;	//boundary handling...

        //y-axis filter is present...
        VOXEL *wrk2 = wrk;

#ifdef I3D_DEBUG
        std::cout << "GaussIIR: convolving in y-axis...\n";
#endif
#ifdef I3D_DEBUG
	std::cout << "GaussIIR: B=" << B << ", b1=" << b1 << ", b2=" << b2 << ", b3=" << b3 << std::endl;
#endif
	//storage of original input values before any y-axis related convolution
	VOXEL *bnd=new VOXEL[x_size];

        for (signed long slices = 0; slices < z_size; slices++)
        {
            wrk = wrk2 + (slices * s_size);

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Y
	    std::cout << "GaussIIR:   forward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //boundary handling...

	    	*(wrk+c)=*(wrk+c) - *(bnd+c)*(b1+b2+b3); //i=0
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += x_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-x_size)*b1 - *(bnd+c)*(b2+b3); //i=1
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += x_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-x_size)*b1 - *(wrk+c-x_2size)*b2 - *(bnd+c)*b3; //i=2
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += x_size;

#ifdef DEBUG_CORE_Y
	    std::cout << "convolutions of i=0,1,2 (slice=" << slices << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < (y_size-1); i++)
            {
#ifdef DEBUG_CORE_Y
                std::cout << "GaussIIR:   forward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c)  - *(wrk+c-x_size)*b1 - *(wrk+c-x_2size)*b2 - *(wrk+c-x_3size)*b3;
		    REMOVE_REMAINDER_FORWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Y
                std::cout << "convolution of " << i << " (slice=" << slices << ") is finished\n";
#endif
                wrk += x_size;
            }

            wrk += x_size;
        }

	wrk2=img.GetVoxelAddr(img.GetImageSize()) -x_size; //important

	VOXEL *bnd2=new VOXEL[x_size];
	VOXEL *bnd3=new VOXEL[x_size];
	VOXEL *vp=new VOXEL[x_size]; //to store v_plus in (15) in [Triggs and Sdika, 2006]

	VOXEL *Bnd2=new VOXEL[x_size];
	VOXEL *Bnd3=new VOXEL[x_size];

        for (signed long slices = 0; slices < z_size; slices++)
        {
            wrk = wrk2 - (slices * s_size);

	    //copy the voxel's value at the end of y-line and finish the forward convolution
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //will be used *bnd instead of *up, just for this occassion

		//finishes the forward convolution
		*(wrk+c)=*(wrk+c) - *(wrk+c-x_size)*b1 - *(wrk+c-x_2size)*b2 - *(wrk+c-x_3size)*b3;
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }

	    for (c = 0; c < x_size; c++) *(vp+c)=*(bnd+c)/denom;
	    for (c = 0; c < x_size; c++) *(bnd3+c)=*(wrk+c-x_2size)-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd2+c)=*(wrk+c-x_size )-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd +c)=*(wrk+c        )-*(bnd+c);

	    //for (c = 0; c < x_size; c++) *(wrk +c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd2+c)=( *(bnd+c)*M[3] + *(bnd2+c)*M[4] + *(bnd3+c)*M[5] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd3+c)=( *(bnd+c)*M[6] + *(bnd2+c)*M[7] + *(bnd3+c)*M[8] + *(vp+c) ) *B;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Y
	    std::cout << "GaussIIR:   backward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(wrk+c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= x_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+x_size)*b1 - *(Bnd2+c)*b2 - *(Bnd3+c)*b3; //i=1
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= x_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+x_size)*b1 - *(wrk+c+x_2size)*b2 - *(Bnd2+c)*b3; //i=2
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= x_size;

#ifdef DEBUG_CORE_Y
	    std::cout << "convolutions of i=" << y_size-1 << "," << y_size-2 << "," << y_size-3 \
	    	      << " (slice=" << slices << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < y_size; i++)
            {
#ifdef DEBUG_CORE_Y
                std::cout << "GaussIIR:   backward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c) * B  - *(wrk+c+x_size)*b1 - *(wrk+c+x_2size)*b2 - *(wrk+c+x_3size)*b3;
	            REMOVE_REMAINDER_BACKWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Y
                std::cout << "convolution of " << y_size-1-i << " (slice=" << slices << ") is finished\n";
#endif
                wrk -= x_size;
            }
        }

	//release allocated memory
	delete[] bnd;
	delete[] bnd2;
	delete[] bnd3;
	delete[] vp;
	delete[] Bnd2;
	delete[] Bnd3;

	wrk=img.GetFirstVoxelAddr(); //important
    }

    // ---- Z AXIS ----
    if ((img.GetSizeZ() > 1) && (SigmaZ >= 1.0))
    {
	double S=static_cast<double>(SigmaZ);
	ComputeBsFromSigma(S,BD,b1D,b2D,b3D);

	b1=static_cast<VOXEL>(b1D);
	b2=static_cast<VOXEL>(b2D);
	b3=static_cast<VOXEL>(b3D);
	B=static_cast<VOXEL>(BD);

	SuggestIIRBoundaries(-b1,-b2,-b3,M);
	const VOXEL denom=1+b1+b2+b3;	//boundary handling...

        //z-axis filter is present...
        VOXEL *wrk2 = wrk;

#ifdef I3D_DEBUG
        std::cout << "GaussIIR: convolving in z-axis...\n";
#endif
#ifdef I3D_DEBUG
	std::cout << "GaussIIR: B=" << B << ", b1=" << b1 << ", b2=" << b2 << ", b3=" << b3 << std::endl;
#endif
	//storage of original input values before any y-axis related convolution
	VOXEL *bnd=new VOXEL[x_size];

        for (signed long cols = 0; cols < y_size; cols++)
        {
            wrk = wrk2 + (cols * x_size);

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Z
	    std::cout << "GaussIIR:   forward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //boundary handling...

	    	*(wrk+c)=*(wrk+c) - *(bnd+c)*(b1+b2+b3); //i=0
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += s_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-s_size)*b1 - *(bnd+c)*(b2+b3); //i=1
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += s_size;

	    for (c = 0; c < x_size; c++) {
	    	*(wrk+c)=*(wrk+c)  - *(wrk+c-s_size)*b1 - *(wrk+c-s_2size)*b2 - *(bnd+c)*b3; //i=2
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }
	    wrk += s_size;

#ifdef DEBUG_CORE_Z
	    std::cout << "convolutions of i=0,1,2 (col=" << cols << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < (z_size-1); i++)
            {
#ifdef DEBUG_CORE_Z
                std::cout << "GaussIIR:   forward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c)  - *(wrk+c-s_size)*b1 - *(wrk+c-s_2size)*b2 - *(wrk+c-s_3size)*b3;
		    REMOVE_REMAINDER_FORWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Z
                std::cout << "convolution of " << i << " (col=" << cols << ") is finished\n";
#endif
                wrk += s_size;
            }

            wrk += s_size;
        }

	wrk2=img.GetVoxelAddr(img.GetImageSize()) -s_size;

	VOXEL *bnd2=new VOXEL[x_size];
	VOXEL *bnd3=new VOXEL[x_size];
	VOXEL *vp=new VOXEL[x_size]; //to store v_plus in (15) in [Triggs and Sdika, 2006]

	VOXEL *Bnd2=new VOXEL[x_size];
	VOXEL *Bnd3=new VOXEL[x_size];

        for (signed long cols = 0; cols < y_size; cols++)
        {
            wrk = wrk2 + (cols * x_size);

	    //copy the voxel's value at the end of z-line and finish the forward convolution
	    for (c = 0; c < x_size; c++) {
		*(bnd+c)=*(wrk+c)/denom; //will be used *bnd instead of *up, just for this occassion

		//finishes the forward convolution
		*(wrk+c)=*(wrk+c) - *(wrk+c-s_size)*b1 - *(wrk+c-s_2size)*b2 - *(wrk+c-s_3size)*b3;
		REMOVE_REMAINDER_FORWARD(*(wrk+c))
	    }

	    for (c = 0; c < x_size; c++) *(vp+c)=*(bnd+c)/denom;
	    for (c = 0; c < x_size; c++) *(bnd3+c)=*(wrk+c-s_2size)-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd2+c)=*(wrk+c-s_size )-*(bnd+c);
	    for (c = 0; c < x_size; c++) *(bnd +c)=*(wrk+c        )-*(bnd+c);

	    //for (c = 0; c < x_size; c++) *(wrk +c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd2+c)=( *(bnd+c)*M[3] + *(bnd2+c)*M[4] + *(bnd3+c)*M[5] + *(vp+c) ) *B;
	    for (c = 0; c < x_size; c++) *(Bnd3+c)=( *(bnd+c)*M[6] + *(bnd2+c)*M[7] + *(bnd3+c)*M[8] + *(vp+c) ) *B;

            // PREFIX, we do it "manually" (aka loop-unrolling)
#ifdef DEBUG_CORE_Z
	    std::cout << "GaussIIR:   backward: prefix starts, ";
#endif
	    for (c = 0; c < x_size; c++) {
		*(wrk+c)=( *(bnd+c)*M[0] + *(bnd2+c)*M[1] + *(bnd3+c)*M[2] + *(vp+c) ) *B;
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= s_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+s_size)*b1 - *(Bnd2+c)*b2 - *(Bnd3+c)*b3; //i=1
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= s_size;

	    for (c = 0; c < x_size; c++) {
	        *(wrk+c)=*(wrk+c)*B  - *(wrk+c+s_size)*b1 - *(wrk+c+s_2size)*b2 - *(Bnd2+c)*b3; //i=2
	        REMOVE_REMAINDER_BACKWARD(*(wrk+c))
	    }
	    wrk -= s_size;

#ifdef DEBUG_CORE_Z
	    std::cout << "convolutions of i=" << z_size-1 << "," << z_size-2 << "," << z_size-3 \
	    	      << " (col=" << cols << ") is finished\n";
#endif
            // INFIX
            for (i = 3; i < z_size; i++)
            {
#ifdef DEBUG_CORE_Z
                std::cout << "GaussIIR:   backward: infix starts, ";
#endif
		for (c = 0; c < x_size; c++) {
		    *(wrk+c)=*(wrk+c) * B  - *(wrk+c+s_size)*b1 - *(wrk+c+s_2size)*b2 - *(wrk+c+s_3size)*b3;
	            REMOVE_REMAINDER_BACKWARD(*(wrk+c))
		}
#ifdef DEBUG_CORE_Z
                std::cout << "convolution of " << z_size-1-i << " (col=" << cols << ") is finished\n";
#endif
                wrk -= s_size;
            }
        }

	//release allocated memory
	delete[] bnd;
	delete[] bnd2;
	delete[] bnd3;
	delete[] vp;
	delete[] Bnd2;
	delete[] Bnd3;
    }

    delete[] M;				//release boundary handling matrix
}

template <class VOXEL, class PRECISION> void GrayGaussIIR(Image3d<VOXEL> &img, const PRECISION SigmaX, const PRECISION SigmaY, const PRECISION SigmaZ) 
{
	Image3d<PRECISION> tmp;
	GrayToFloat(img, tmp);
	GaussIIR(tmp, SigmaX, SigmaY, SigmaZ);
	FloatToGrayNoWeight(tmp, img);
}

// ---------------------------
// Vlado's IIR Gauss ends here
// ---------------------------

// ----------------------------------------
// Vlado's SeparableConvolution starts here
// ----------------------------------------

// --------- struct Separable3DFilter ---------
template <class VOXEL> Separable3dFilter<VOXEL>::Separable3dFilter(struct Separable3dFilter<VOXEL> const &f, int mirror)
{
    // safeback init...
    xData=yData=zData=NULL;
    xLength=yLength=zLength=0;
    xDiv=yDiv=zDiv=1;

    VOXEL **tmp = new VOXEL *[3];
    tmp[0] = f.xData;
    tmp[1] = f.yData;
    tmp[2] = f.zData;
    int i, j;

    // "reduce" the number of pointers
    for (i = 0; i < 2; ++i)
        for (j = i + 1; j < 3; ++j)
            if (tmp[i] == tmp[j]) tmp[j] = NULL;

    // (mirror-)copy relevant data
    if (tmp[0] != NULL)
    {
        xData=new VOXEL[f.xLength];
        xLength=f.xLength; xDiv=f.xDiv;

	if (mirror) { for (i=0; i < xLength; ++i) xData[i]=f.xData[xLength-1 -i]; }
	else { for (i=0; i < xLength; ++i) xData[i]=f.xData[i]; }
    }
    if (tmp[1] != NULL)
    {
        yData=new VOXEL[f.yLength];
        yLength=f.yLength; yDiv=f.yDiv;

	if (mirror) { for (i=0; i < yLength; ++i) yData[i]=f.yData[yLength-1 -i]; }
	else { for (i=0; i < yLength; ++i) yData[i]=f.yData[i]; }
    }
    if (tmp[2] != NULL)
    {
        zData=new VOXEL[f.zLength];
        zLength=f.zLength; zDiv=f.zDiv;

	if (mirror) { for (i=0; i < zLength; ++i) zData[i]=f.zData[zLength-1 -i]; }
	else { for (i=0; i < zLength; ++i) zData[i]=f.zData[i]; }
    }

    delete[]tmp;

    // "bump-up back" the number of pointers
    if (f.yData == f.xData) { yData=xData; yLength=xLength; yDiv=xDiv; }
    if (f.zData == f.yData) { zData=yData; zLength=yLength; zDiv=yDiv; }
    else if (f.zData == f.xData) { zData=xData; zLength=xLength; zDiv=xDiv; }
}

template <class VOXEL> void Separable3dFilter<VOXEL>::PrintFilter(void) {
	std::cout << "3D filter with the following content:\nX: ";
	if (xLength > 0) {
		std::cout << xLength << " values: ";
		for (int i=0; i < (xLength-1); ++i) std::cout << xData[i] << ",";
		std::cout << xData[xLength-1] << " @" << reinterpret_cast<long>(xData) << ", ";
		std::cout << "divisor is " << xDiv << "\nY: ";

	} else std::cout << "empty\nY: ";
	if (yLength > 0) {
		std::cout << yLength << " values: ";
		for (int i=0; i < (yLength-1); ++i) std::cout << yData[i] << ",";
		std::cout << yData[yLength-1] << " @" << reinterpret_cast<long>(yData) << ", ";
		std::cout << "divisor is " << yDiv << "\nZ: ";

	} else std::cout << "empty\nZ: ";
	if (zLength > 0) {
		std::cout << zLength << " values: ";
		for (int i=0; i < (zLength-1); ++i) std::cout << zData[i] << ",";
		std::cout << zData[zLength-1] << " @" << reinterpret_cast<long>(zData) << ", ";
		std::cout << "divisor is " << zDiv << "\n\n";

	} else std::cout << "empty\n\n";
}

template <class VOXEL> void Separable3dFilter<VOXEL>::DisposeData(void)
{
    VOXEL **tmp = new VOXEL *[3];
    tmp[0] = xData;
    tmp[1] = yData;
    tmp[2] = zData;
    int i, j, imax = 3;

    for (i = 0; i < imax - 1; ++i)
        for (j = i + 1; j < imax; ++j)
        {
            if (tmp[i] == tmp[j])
                tmp[j] = NULL;
        }

    for (i = 0; i < imax; ++i)
    {
        if (tmp[i] != NULL)
            delete[]tmp[i];
        tmp[i] = NULL;
    }
    delete[]tmp;
}

template <class VOXEL> Separable3dFilter<VOXEL>::~Separable3dFilter()
{
	this->DisposeData();
}


// --------- struct BorderPadding ---------
template <class VOXEL> BorderPadding<VOXEL>::BorderPadding(struct BorderPadding<VOXEL> const &f, int mirror)
{
    // safeback init...
    xDataPre=yDataPre=zDataPre=NULL;
    xDataPost=yDataPost=zDataPost=NULL;
    xLength=yLength=zLength=0;

    VOXEL **tmp = new VOXEL *[6];
    tmp[0] = f.xDataPre;
    tmp[1] = f.xDataPost;
    tmp[2] = f.yDataPre;
    tmp[3] = f.yDataPost;
    tmp[4] = f.zDataPre;
    tmp[5] = f.zDataPost;
    int i, j;

    // "reduce" the number of pointers
    for (i = 0; i < 5; ++i)
        for (j = i + 1; j < 6; ++j)
            if (tmp[i] == tmp[j]) tmp[j] = NULL;

    // (mirror-)copy relevant data
    if (tmp[0] != NULL)
    {
        xDataPre=new VOXEL[f.xLength];
        if (f.xDataPre != f.xDataPost) xDataPost=new VOXEL[f.xLength]; else xDataPost=xDataPre;
        xLength=f.xLength;

	if (mirror) {
		for (i=0; i < xLength; ++i) xDataPre[i]=f.xDataPre[xLength-1 -i];
		if (f.xDataPre != f.xDataPost) for (i=0; i < xLength; ++i) xDataPost[i]=f.xDataPost[xLength-1 -i];
	} else {
		for (i=0; i < xLength; ++i) xDataPre[i]=f.xDataPre[i];
		if (f.xDataPre != f.xDataPost) for (i=0; i < xLength; ++i) xDataPost[i]=f.xDataPost[i];
	}
    }
    if (tmp[2] != NULL)
    {
        yDataPre=new VOXEL[f.yLength];
        if (f.yDataPre != f.yDataPost) yDataPost=new VOXEL[f.yLength]; else yDataPost=yDataPre;
        yLength=f.yLength;

	if (mirror) {
		for (i=0; i < yLength; ++i) yDataPre[i]=f.yDataPre[yLength-1 -i];
		if (f.yDataPre != f.yDataPost) for (i=0; i < xLength; ++i) yDataPost[i]=f.yDataPost[xLength-1 -i];
	} else {
		for (i=0; i < yLength; ++i) yDataPre[i]=f.yDataPre[i];
		if (f.yDataPre != f.yDataPost) for (i=0; i < xLength; ++i) yDataPost[i]=f.yDataPost[i];
	}
    }
    if (tmp[4] != NULL)
    {
        zDataPre=new VOXEL[f.zLength];
        if (f.zDataPre != f.zDataPost) zDataPost=new VOXEL[f.zLength]; else zDataPost=zDataPre;
        zLength=f.zLength;

	if (mirror) {
		for (i=0; i < zLength; ++i) zDataPre[i]=f.zDataPre[zLength-1 -i];
		if (f.zDataPre != f.zDataPost) for (i=0; i < xLength; ++i) zDataPost[i]=f.zDataPost[xLength-1 -i];
	} else {
		for (i=0; i < zLength; ++i) zDataPre[i]=f.zDataPre[i];
		if (f.zDataPre != f.zDataPost) for (i=0; i < xLength; ++i) zDataPost[i]=f.zDataPost[i];
	}
    }

    delete[]tmp;

    // "bump-up back" the number of pointers
    if (f.yDataPre == f.xDataPre) { yDataPre=xDataPre; yDataPost=xDataPost; yLength=xLength; }
    if (f.zDataPre == f.yDataPre) { zDataPre=yDataPre; zDataPost=yDataPost; zLength=yLength; }
    else if (f.zDataPre == f.xDataPre) { zDataPre=xDataPre; zDataPost=xDataPost; zLength=xLength; }
}

template <class VOXEL> void BorderPadding<VOXEL>::PrintPaddings(void) {
	std::cout << "Border paddings with the following content:\nX: ";
	if (xLength > 0) {
		std::cout << xLength << " values: ";
		for (int i=0; i < (xLength-1); ++i) std::cout << xDataPre[i] << ",";
		std::cout << xDataPre[xLength-1] << " @" << reinterpret_cast<long>(xDataPre) << " <-> ";
		for (int i=0; i < (xLength-1); ++i) std::cout << xDataPost[i] << ",";
		std::cout << xDataPost[xLength-1] << " @" << reinterpret_cast<long>(xDataPost) << "\nY: ";

	} else std::cout << "empty\nY: ";
	if (yLength > 0) {
		std::cout << yLength << " values: ";
		for (int i=0; i < (yLength-1); ++i) std::cout << yDataPre[i] << ",";
		std::cout << yDataPre[yLength-1] << " @" << reinterpret_cast<long>(yDataPre) << " <-> ";
		for (int i=0; i < (yLength-1); ++i) std::cout << yDataPost[i] << ",";
		std::cout << yDataPost[yLength-1] << " @" << reinterpret_cast<long>(yDataPost) << "\nZ: ";

	} else std::cout << "empty\nZ: ";
	if (zLength > 0) {
		std::cout << zLength << " values: ";
		for (int i=0; i < (zLength-1); ++i) std::cout << zDataPre[i] << ",";
		std::cout << zDataPre[zLength-1] << " @" << reinterpret_cast<long>(zDataPre) << " <-> ";
		for (int i=0; i < (zLength-1); ++i) std::cout << zDataPost[i] << ",";
		std::cout << zDataPost[zLength-1] << " @" << reinterpret_cast<long>(zDataPost) << "\n\n";

	} else std::cout << "empty\n\n";
}

template <class VOXEL> void BorderPadding<VOXEL>::DisposeData(void)
{
    VOXEL **tmp = new VOXEL *[6];
    tmp[0] = xDataPre;
    tmp[1] = xDataPost;
    tmp[2] = yDataPre;
    tmp[3] = yDataPost;
    tmp[4] = zDataPre;
    tmp[5] = zDataPost;

    int i, j, imax = 6;

    for (i = 0; i < imax - 1; ++i)
        for (j = i + 1; j < imax; ++j)
        {
            if (tmp[i] == tmp[j])
                tmp[j] = NULL;
        }

    for (i = 0; i < imax; ++i)
    {
        if (tmp[i] != NULL)
            delete[]tmp[i];
        tmp[i] = NULL;
    }
    delete[]tmp;
}

template <class VOXEL> BorderPadding<VOXEL>::~BorderPadding()
{
	this->DisposeData();
}


// --------- convolution function ---------
template <class VOXEL>
  int SeparableConvolution(i3d::Image3d<VOXEL> const &d, 
				 i3d::Image3d<VOXEL> &res,
				 struct Separable3dFilter<VOXEL> const &f,
				 struct BorderPadding<VOXEL> const &b)
{
    //first of all, a few integrity tests:
    //filter sizes must be odd or zero
    if (!(f.xLength & 1) && (f.xLength != 0))
    {
        std::cerr << "SeparableConvolution(): x-axis filter length is wrong!\n";
        return (1);
    }
    if (!(f.yLength & 1) && (f.yLength != 0))
    {
        std::cerr << "SeparableConvolution(): y-axis filter length is wrong!\n";
        return (1);
    }
    if (!(f.zLength & 1) && (f.zLength != 0))
    {
        std::cerr << "SeparableConvolution(): z-axis filter length is wrong!\n";
        return (1);
    }
#ifdef I3D_DEBUG
    std::cout << "SeparableConvolution(): filter sizes tests passed\n";
#endif

    //padding data length over both sides of given axis must be
    //exactly the filter length minus 1 for the given axis
    if ((2 * b.xLength != f.xLength - 1) && (f.xLength != 0))
    {

        std::cerr << "SeparableConvolution(): Padding in x-axis is insufficient!\n";
        return (2);
    }
    if ((2 * b.yLength != f.yLength - 1) && (f.yLength != 0))
    {
        std::cerr << "SeparableConvolution(): Padding in y-axis is insufficient!\n";
        return (2);
    }
    if ((2 * b.zLength != f.zLength - 1) && (f.zLength != 0))
    {
        std::cerr << "SeparableConvolution(): Padding in z-axis is insufficient!\n";
        return (2);
    }
#ifdef I3D_DEBUG
    std::cout << "SeparableConvolution(): border sizes vs filter sizes tests passed\n";
#endif

    //input/output image dimensions
    signed long x_size = (signed long) d.GetWidth();
    signed long y_size = (signed long) d.GetHeight();
    signed long z_size = (signed long) d.GetNumSlices();

    //determine in which image to store the intermediate results first: res or tmp?
    short output_selector=0;
    output_selector+=(f.xLength > 0)? 1: 0;
    output_selector+=(f.yLength > 0)? 1: 0;
    output_selector+=(f.zLength > 0)? 1: 0;

    res.SetOffset(d.GetOffset());
    res.SetResolution(d.GetResolution());
    if (output_selector > 0) res.MakeRoom(d.GetSize());

    //an alternating output image and its metadata
    Image3d <VOXEL> tmp;
    tmp.SetOffset(d.GetOffset());
    tmp.SetResolution(d.GetResolution());
    if (output_selector > 1) tmp.MakeRoom(d.GetSize());

#ifdef I3D_DEBUG
    std::cout << "SeparableConvolution(): input image of size " << x_size << "x" << y_size << "x" << z_size << " voxels\n";
    std::cout << "SeparableConvolution(): borders are " << b.xLength << ", " << b.yLength << ", " << b.zLength << " voxels wide\n";
    std::cout << "SeparableConvolution(): filter sizes are " << f.xLength << ", " << f.yLength << ", " << f.zLength << " voxels\n";
#endif

    //temporary variables
    signed long i = 0, j = 0, k = 0, c = 0;     //,c2=0;

    //in points to memory from where input for current convolution direction goes in
    //out points to memory where output for current convolution direction should go
    const VOXEL *in = d.GetFirstVoxelAddr();    //the very first input data
    VOXEL *out = res.GetFirstVoxelAddr();
    if (output_selector == 2) {
    	out=tmp.GetFirstVoxelAddr();
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution(): input set on the original input image\n";
        std::cout << "SeparableConvolution(): output set to the temporary image\n";
#endif
    } else {
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution(): input set on the original input image\n";
        std::cout << "SeparableConvolution(): output set to the output image\n";
#endif
    }

    // ---- X AXIS ----
    if (f.xLength > 0)
    {
        //x-axis filter is present...
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution(): convolving in x-axis...\n";
#endif
        //since prefix and postfix convolution code works with just constants,
        //we can precompute the results of all partial convolutions; this saves
        //some time in the main loop
        VOXEL *conv_pre = new VOXEL[b.xLength + 1];
        VOXEL *conv_post = new VOXEL[b.xLength + 1];
        if ((!conv_pre) || (!conv_post))
        {
            std::cerr << "SeparableConvolution(): No memory to store convolution constants!\n";

            if (conv_pre)
                delete[]conv_pre;
            if (conv_post)
                delete[]conv_post;
            return (3);
        }
        //initiate memory
        for (j = 0; j <= b.xLength; j++)
        {
            conv_pre[j] = 0;
            conv_post[j] = 0;
        }

        //compute partial convolutions
        for (j = 0; j < b.xLength; j++)
            for (k = 0; k < b.xLength - j; k++)
            {
                conv_pre[b.xLength - j] += b.xDataPre[k + j] * f.xData[k];
                conv_post[b.xLength - j] += b.xDataPost[k + j] * f.xData[b.xLength + 1 + k];
            }
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution():   partial convolutions precomputed\n";
#endif
        //for each row...
        signed long min = (b.xLength < x_size) ? b.xLength : x_size;
        signed long max = ((x_size - b.xLength) > b.xLength) ? (x_size - b.xLength) : b.xLength;

        for (signed long rows = 0; rows < y_size * z_size; rows++)
        {
            //no need to shift in and out pointers because
            //x-lines are sequentially in memory

            //j relative position, k position in filter

            // PREFIX
            for (i = 0; i < min; i++)
            {
#ifdef DEBUG_CORE_X
                std::cout << "SeparableConvolution(): PRE: prefix starts (k=0), ";
#endif
                //prefix
                k = b.xLength - i;
                *out = conv_pre[k];
#ifdef DEBUG_CORE_X
                std::cout << "infix starts (k=" << k << "), ";
#endif
                //infix
                for (j = -b.xLength + k; (i + j < x_size) && (k < f.xLength); j++, k++)
                    *out += *(in + j) * f.xData[k];
#ifdef DEBUG_CORE_X
                std::cout << "postfix starts (k=" << k << "), ";
#endif
                //postfix
                k = (k < f.xLength) ? f.xLength - k : 0;
                *out += conv_post[k];

                //this is better to compute afterwards, after all summings
                *out /= f.xDiv;
#ifdef DEBUG_CORE_X
                std::cout << "convolution of " << i << " (row=" << rows << ") is finished\n";
#endif
                in++;
                out++;
            }

            // INFIX
            for (i = b.xLength; i < (x_size - b.xLength); i++)
            {
                k = 0;
                *out = 0;
#ifdef DEBUG_CORE_X
                std::cout << "SeparableConvolution(): IN: infix starts (k=0), ";
#endif
                //infix
                for (j = -b.xLength; k < f.xLength; j++, k++)
                    *out += *(in + j) * f.xData[k];

                //this is better to compute afterwards, after all summings
                *out /= f.xDiv;
#ifdef DEBUG_CORE_X
                std::cout << "convolution of " << i << " (row=" << rows << ") is finished\n";
#endif
                in++;
                out++;
            }

            // POSTFIX
            for (i = max; i < x_size; i++)
            {
                k = 0;
                *out = 0;
#ifdef DEBUG_CORE_X
                std::cout << "SeparableConvolution(): POST: infix starts (k=0), ";
#endif
                //infix
                for (j = -b.xLength; i + j < x_size; j++, k++)
                    *out += *(in + j) * f.xData[k];
#ifdef DEBUG_CORE_X
                std::cout << "postfix starts (k=" << k << "), ";
#endif
                //postfix
                *out += conv_post[f.xLength - k];

                //this is better to compute afterwards, after all summings
                *out /= f.xDiv;
#ifdef DEBUG_CORE_X
                std::cout << "convolution of " << i << " (row=" << rows << ") is finished\n";
#endif
                in++;
                out++;
            }
        }

	if (output_selector == 2) {
	    //we were storing into tmp
	    in=tmp.GetFirstVoxelAddr();
	    out=res.GetFirstVoxelAddr();
#ifdef I3D_DEBUG
            std::cout << "SeparableConvolution():   input set on the temporary image\n";
            std::cout << "SeparableConvolution():   output set to the output image\n";
#endif
	} else {
	    //we were storing into res
	    //note: there might be no other convolution taking place (output_selector == 1)
	    in=res.GetFirstVoxelAddr();
	    out=tmp.GetFirstVoxelAddr();
#ifdef I3D_DEBUG
            std::cout << "SeparableConvolution():   input set on the output image\n";
            std::cout << "SeparableConvolution():   output set to the temporary image\n";
#endif
	}
    	--output_selector;

        delete[]conv_pre;
        delete[]conv_post;
        conv_pre = conv_post = NULL;
    }

    // ---- Y AXIS ----
    if (f.yLength > 0)
    {
        //y-axis filter is present...
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution(): convolving in y-axis...\n";
#endif
        //since prefix and postfix convolution code works with just constants,
        //we can precompute the results of all partial convolutions; this saves
        //some time in the main loop
        VOXEL *conv_pre = new VOXEL[b.yLength + 1];
        VOXEL *conv_post = new VOXEL[b.yLength + 1];
        if ((!conv_pre) || (!conv_post))
        {
            std::cerr << "SeparableConvolution(): No memory to store convolution constants!\n";

            if (conv_pre)
                delete[]conv_pre;
            if (conv_post)
                delete[]conv_post;
            return (3);
        }
        //initiate memory
        for (j = 0; j <= b.yLength; j++)
        {
            conv_pre[j] = 0;
            conv_post[j] = 0;
        }

        //compute partial convolutions
        for (j = 0; j < b.yLength; j++)
            for (k = 0; k < b.yLength - j; k++)
            {
                conv_pre[b.yLength - j] += b.yDataPre[k + j] * f.yData[k];
                conv_post[b.yLength - j] += b.yDataPost[k + j] * f.yData[b.yLength + 1 + k];
            }
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution():   partial convolutions precomputed\n";
#endif
        //we don't know the image to which the in pointer belongs, we need to copy it;
        //this is required for setting correct offsets of each column
        const VOXEL *in2 = in;
        VOXEL *out2 = out;

        //for each column...
        signed long min = (b.yLength < y_size) ? b.yLength : y_size;
        signed long max = ((y_size - b.yLength) > b.yLength) ? (y_size - b.yLength) : b.yLength;

        for (signed long slices = 0; slices < z_size; slices++)
        {
            in = in2 + (slices * y_size * x_size);
            out = out2 + (slices * y_size * x_size);

            //j relative position, k position in filter

            // PREFIX
            for (i = 0; i < min; i++)
            {
#ifdef DEBUG_CORE_Y
                std::cout << "SeparableConvolution(): PRE: prefix starts (k=0), ";
#endif
                //prefix
                k = b.yLength - i;
                //must go, since it initiates output memory
                for (c = 0; c < x_size; c++)
                    *(out + c) = conv_pre[k];
#ifdef DEBUG_CORE_Y
                std::cout << "infix starts (k=" << k << "), ";
#endif
                //infix
                for (j = -b.yLength + k; (i + j < y_size) && (k < f.yLength); j++, k++)
                    for (c = 0; c < x_size; c++)
                        *(out + c) += *(in + (j * x_size) + c) * f.yData[k];
#ifdef DEBUG_CORE_Y
                std::cout << "postfix starts (k=" << k << "), ";
#endif
                //postfix, the last two steps together...
                k = (k < f.yLength) ? f.yLength - k : 0;
                for (c = 0; c < x_size; c++)
                    *(out + c) = (*(out + c) + conv_post[k]) / f.yDiv;
#ifdef DEBUG_CORE_Y
                std::cout << "convolution of " << i << " (slice=" << slices << ") is finished\n";
#endif
                in += x_size;
                out += x_size;
            }

            // INFIX
            for (i = b.yLength; i < (y_size - b.yLength); i++)
            {
                k = 0;
                //must go, since it initiates output memory
                for (c = 0; c < x_size; c++)
                    *(out + c) = 0;
#ifdef DEBUG_CORE_Y
                std::cout << "SeparableConvolution(): IN: infix starts (k=0), ";
#endif
                //infix
                for (j = -b.yLength; k < f.yLength; j++, k++)
                    for (c = 0; c < x_size; c++)
                        *(out + c) += *(in + (j * x_size) + c) * f.yData[k];

                //this is better to compute afterwards, after all summings
                for (c = 0; c < x_size; c++)
                    *(out + c) /= f.yDiv;
#ifdef DEBUG_CORE_Y
                std::cout << "convolution of " << i << " (slice=" << slices << ") is finished\n";
#endif
                in += x_size;
                out += x_size;
            }

            // POSTFIX
            for (i = max; i < y_size; i++)
            {
                k = 0;
                //must go, since it initiates output memory
                for (c = 0; c < x_size; c++)
                    *(out + c) = 0;
#ifdef DEBUG_CORE_Y
                std::cout << "SeparableConvolution(): POST: infix starts (k=0), ";
#endif
                //infix
                for (j = -b.yLength; i + j < y_size; j++, k++)
                    for (c = 0; c < x_size; c++)
                        *(out + c) += *(in + (j * x_size) + c) * f.yData[k];
#ifdef DEBUG_CORE_Y
                std::cout << "postfix starts (k=" << k << "), ";
#endif
                //postfix, the last two steps together...
                for (c = 0; c < x_size; c++)
                    *(out + c) = (*(out + c) + conv_post[f.yLength - k]) / f.yDiv;
#ifdef DEBUG_CORE_Y
                std::cout << "convolution of " << i << " (slice=" << slices << ") is finished\n";
#endif
                in += x_size;
                out += x_size;
            }
        }

	in=out2; //that's the only sure thing
	if (out2 == tmp.GetFirstVoxelAddr()) {
	    out=res.GetFirstVoxelAddr();
#ifdef I3D_DEBUG
            std::cout << "SeparableConvolution():   input set to the temporary image\n";
            std::cout << "SeparableConvolution():   output set to the output image\n";
#endif
	} else {
	    out=tmp.GetFirstVoxelAddr();
#ifdef I3D_DEBUG
            std::cout << "SeparableConvolution():   input set to the output image\n";
            std::cout << "SeparableConvolution():   output set to the temporary image\n";
#endif
	}
    	--output_selector;

        delete[]conv_pre;
        delete[]conv_post;
        conv_pre = conv_post = NULL;
    }

    // ---- Z AXIS ----
    if (f.zLength > 0)
    {
        //z-axis filter is present...
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution(): convolving in z-axis...\n";
#endif
        //since prefix and postfix convolution code works with just constants,
        //we can precompute the results of all partial convolutions; this saves
        //some time in the main loop
        VOXEL *conv_pre = new VOXEL[b.zLength + 1];
        VOXEL *conv_post = new VOXEL[b.zLength + 1];
        if ((!conv_pre) || (!conv_post))
        {
            std::cerr << "SeparableConvolution(): No memory to store convolution constants!\n";

            if (conv_pre)
                delete[]conv_pre;
            if (conv_post)
                delete[]conv_post;
            return (3);
        }
        //initiate memory
        for (j = 0; j <= b.zLength; j++)
        {
            conv_pre[j] = 0;
            conv_post[j] = 0;
        }

        //compute partial convolutions
        for (j = 0; j < b.zLength; j++)
            for (k = 0; k < b.zLength - j; k++)
            {
                conv_pre[b.zLength - j] += b.zDataPre[k + j] * f.zData[k];
                conv_post[b.zLength - j] += b.zDataPost[k + j] * f.zData[b.zLength + 1 + k];
            }
#ifdef I3D_DEBUG
        std::cout << "SeparableConvolution():   partial convolutions precomputed\n";
#endif
        //we don't know the image to which the in pointer belongs, we need to copy it;
        //this is required for setting correct offsets of each column
        const VOXEL *in2 = in;
        VOXEL *out2 = out;

        const signed long s_size = x_size * y_size;     //just helping variable

        //for each slice...
        signed long min = (b.zLength < z_size) ? b.zLength : z_size;
        signed long max = ((z_size - b.zLength) > b.zLength) ? (z_size - b.zLength) : b.zLength;

        for (signed long cols = 0; cols < y_size; cols++)
        {
            in = in2 + (cols * x_size);
            out = out2 + (cols * x_size);

            //j relative position, k position in filter

            // PREFIX
            for (i = 0; i < min; i++)
            {
#ifdef DEBUG_CORE_Z
                std::cout << "SeparableConvolution(): PRE: prefix starts (k=0), ";
#endif
                //prefix
                k = b.zLength - i;
                //must go, since it initiates output memory
                for (c = 0; c < x_size; c++)
                    *(out + c) = conv_pre[k];
#ifdef DEBUG_CORE_Z
                std::cout << "infix starts (k=" << k << "), ";
#endif
                //infix
                for (j = -b.zLength + k; (i + j < z_size) && (k < f.zLength); j++, k++)
                    for (c = 0; c < x_size; c++)
                        *(out + c) += *(in + (j * s_size) + c) * f.zData[k];
#ifdef DEBUG_CORE_Z
                std::cout << "postfix starts (k=" << k << "), ";
#endif
                //postfix, the last two steps together...
                k = (k < f.zLength) ? f.zLength - k : 0;
                for (c = 0; c < x_size; c++)
                    *(out + c) = (*(out + c) + conv_post[k]) / f.zDiv;
#ifdef DEBUG_CORE_Z
                std::cout << "convolution of " << i << " (col=" << cols << ") is finished\n";
#endif
                in += s_size;
                out += s_size;
            }

            // INFIX
            for (i = b.zLength; i < (z_size - b.zLength); i++)
            {
                k = 0;
                //must go, since it initiates output memory
                for (c = 0; c < x_size; c++)
                    *(out + c) = 0;
#ifdef DEBUG_CORE_Z
                std::cout << "SeparableConvolution(): IN: infix starts (k=0), ";
#endif
                //infix
                for (j = -b.zLength; k < f.zLength; j++, k++)
                    for (c = 0; c < x_size; c++)
                        *(out + c) += *(in + (j * s_size) + c) * f.zData[k];

                //this is better to compute afterwards, after all summings
                for (c = 0; c < x_size; c++)
                    *(out + c) /= f.zDiv;
#ifdef DEBUG_CORE_Z
                std::cout << "convolution of " << i << " (col=" << cols << ") is finished\n";
#endif
                in += s_size;
                out += s_size;
            }

            // POSTFIX
            for (i = max; i < z_size; i++)
            {
                k = 0;
                //must go, since it initiates output memory
                for (c = 0; c < x_size; c++)
                    *(out + c) = 0;
#ifdef DEBUG_CORE_Z
                std::cout << "SeparableConvolution(): POST: infix starts (k=0), ";
#endif
                //infix
                for (j = -b.zLength; i + j < z_size; j++, k++)
                    for (c = 0; c < x_size; c++)
                        *(out + c) += *(in + (j * s_size) + c) * f.zData[k];
#ifdef DEBUG_CORE_Z
                std::cout << "postfix starts (k=" << k << "), ";
#endif
                //postfix, the last two steps together...
                for (c = 0; c < x_size; c++)
                    *(out + c) = (*(out + c) + conv_post[f.zLength - k]) / f.zDiv;
#ifdef DEBUG_CORE_Z
                std::cout << "convolution of " << i << " (col=" << cols << ") is finished\n";
#endif
                in += s_size;
                out += s_size;
            }
        }

    	--output_selector;
        in = out = NULL;
        delete[]conv_pre;
        delete[]conv_post;
        conv_pre = conv_post = NULL;
    }

    tmp.DisposeData();
    return (0);
}
// --------------------------------------
// Vlado's SeparableConvolution ends here
// --------------------------------------


/***************************************************************************/ 

/* Explicit instantiations: */
template I3D_ALGO_EXPORT void GaussIIR(Image3d<float>&, const float);
template I3D_ALGO_EXPORT void GaussIIR(Image3d<double>&, const double);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY8>&, const float);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY8>&, const double);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY16>&, const float);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY16>&, const double);
template I3D_ALGO_EXPORT void GaussIIR(Image3d<float>&, const float, const float, const float);
template I3D_ALGO_EXPORT void GaussIIR(Image3d<double>&, const double, const double, const double);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY8>&, const float, const float, const float);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY8>&, const double, const double, const double);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY16>&, const float, const float, const float);
template I3D_ALGO_EXPORT void GrayGaussIIR(Image3d<GRAY16>&, const double, const double, const double);

template I3D_ALGO_EXPORT void GaussFIR(Image3d<float> const &, Image3d<float>&, const float, const float);
template I3D_ALGO_EXPORT void GaussFIR(Image3d<double> const &, Image3d<double>&, const double, const float);
template I3D_ALGO_EXPORT void GaussFIR(Image3d<float>&, const float, const float);
template I3D_ALGO_EXPORT void GaussFIR(Image3d<double>&, const double, const float);
template I3D_ALGO_EXPORT void GaussFIR(Image3d<float> const &, Image3d <float>&, const float, const float, const float, const float, const float, const float);
template I3D_ALGO_EXPORT void GaussFIR(Image3d<double> const &, Image3d <double>&, const double, const double, const double, const float, const float, const float);
template I3D_ALGO_EXPORT void GaussFIR(Image3d<float>&, const float, const float, const float, const float, const float, const float);
template I3D_ALGO_EXPORT void GaussFIR(Image3d<double>&, const double, const double, const double, const float, const float, const float);

}
